require "tools"
require "c8y_bin_api"

--
-- SAVE:
--   dbcfg_export -o /tmp/$filename -p "$pw"
--   tar -C /usr/local/cdcs -zcvf - ipsec.d openvpn-keys ssh-hostkeys | openssl des3 -salt -k "$pw" | dd of=/tmp/vpn.des3
--   tar -zcvf $filename.tar.gz $filename vpn.des3
--
--
-- RESTORE:
--   install_file "CONFIG_FILE"    (WARNING: it clears missing variables)
--

local templates = {}
local inProgress = false
local tempFile = c8y.config:value("tempFile", "/opt/cdcs/upload/configFile.cfg.tar.gz")
local credencialsFile = c8y.config:value("tempFile", "/opt/sragent/config/credentials")

local debug = c8y.config:boolValue("debug") and print or function() end

local removeCredentialsSedPattern = "'/service.cumulocity.connection.\\(password\\|username\\)/d'"



-----------------------------------------------------------------------
-- Install configuration
local function installFile(fileName)
	fileName = fileName or tempFile
  debug('INSTALL FILE "' .. fileName .. '"')
	-- save credentials
	os.execute('mkdir -p ' .. dirname(credencialsFile) .. '; rdb_get -L service.cumulocity.connection. > ' .. credencialsFile)
	-- install
  local file = io.popen("install_file " .. fileName)
  local response = file:read("*a")
	debug('INSTALING...', response, '###')
  return response:find("is successful") ~= nil, response
end

-----------------------------------------------------------------------
-- Restart the device
local function reboot()
  debug("REBOOT")
  os.execute("reboot")
end

-----------------------------------------------------------------------
-- Remove credetials from cofig file
local function removeCredentials(fileName, zip)
  local tmpPath, zipFileName
  if zip then
    tmpPath = string.format('/tmp/c%06d', math.random(999999))
    zipFileName, fileName = fileName, tmpPath..'/NTC-6200*.cfg'
    os.execute(string.format('mkdir -p %s; tar -C %s -xzf %s', tmpPath, tmpPath, zipFileName))
  end
  os.execute("sed -i " .. removeCredentialsSedPattern .." " .. fileName)
  if zip then
    os.execute(string.format('tar -C %s -czf %s .; rm -r %s', tmpPath, zipFileName, tmpPath))
  end
end

-----------------------------------------------------------------------
-- Handle download operation
local function download(op)
  if inProgress then
    op:setFailed("Config operation already in progress")
  else
    inProgress = true
    local url = op:getResponseRecord(1)[3]
    local configDumpId = op:getResponseRecord(1)[4]
    debug("EXECUTING DOWNLOAD HANDLER", op.id, url, configDumpId)
		c8y_bin_api.download{
			url = url,
			fileName = tempFile,
			onSuccess = function(fileName)
				removeCredentials(fileName, 'zip')
				if installFile(fileName) then
					templates.updateConfigDumpFragment:send{c8y.device.id, configDumpId}
					reboot()
				else
					debug("Installing the config file failed")
					op:setFailed("Installing the config file failed")
					inProgress = false
				end
			end,
			onError = function(e, errorText)
				debug("Download failed: " .. (errorText or "???"))
				op:setFailed(errorText or "Unknown error")
				inProgress = false
			end
		}
  end
end

-----------------------------------------------------------------------
-- Export config to the file
local function exportConfig()
  local fileName = string.format("NTC-6200_Settings_%d_%s.cfg", c8y.device.id, c8y.strftime('%Y%m%d'))
	local filePath = '/opt/cdcs/upload/' .. fileName .. '.tar.gz'
  if os.execute("dbcfg_export -o /tmp/" .. fileName ..  ' -p ""') ~= 0 then
    return nil, "Exporting rdb config failed"
  end
	removeCredentials('/tmp/' .. fileName)
	local openvpnkeys = file_entry_exists('/usr/local/cdcs/openvpn-keys') and 'openvpn-keys ' or ''
	local sshhostkeys = file_entry_exists('/usr/local/cdcs/ssh-hostkeys') and 'ssh-hostkeys ' or ''
  if  os.execute('tar -C /usr/local/cdcs -zcvf - ipsec.d ' .. openvpnkeys .. sshhostkeys .. ' | openssl des3 -salt -k "" | dd of=/tmp/vpn.des3') ~= 0 then
    return nil, "Exporting vpn config failed"
  end
  if os.execute("cd /tmp/ && tar -zcf " .. filePath .. " " .. fileName .. " vpn.des3") ~= 0 then
    return nil, "Archiving config failed"
  end
	os.remove('/tmp/' .. fileName)
  return filePath
end

-----------------------------------------------------------------------
-- Update current configuration fragment in the device
local function updateConfigDumpFragment(configDumpId)
  local response = templates.updateConfigDumpFragment:send({c8y.device, configDumpId})
  if response.httpCode == 200 then
    return true
	end
	return false, "Error updating configuration dump device fragment"
end

-----------------------------------------------------------------------
-- Handle upload operation
local function upload(op)
  debug("EXECUTING UPLOAD HANDLER", op.id)
  local fileName, errorText = exportConfig()
  if not fileName then
    op:setStatus(op.Failed, errorText)
    return
  end

	c8y_bin_api.upload{
		fileName = fileName,
		onSuccess = function(binaryId)
			local url = c8y_bin_api.url(binaryId)
			local name = "NTC 6200 Device Configuration"
			local descr = string.format("Uploaded by device %d: %s at %s",
				c8y.device.id, c8y.device.defaultName, c8y.strftime('%F %T'))
			local response = templates.createConfigRepoEntry:send{name, descr, url}
			if response.ok then
				local configDumpId = response:getRecord(0, templates.response.id)[3]
				if configDumpId then
					local success, errorText = updateConfigDumpFragment(configDumpId)
					op:setStatus(success and op.Successful or op.Failed, errorText)
					return
				end
			end
			op:setFailed('Creating config repo entry failed')
		end,
		onError = function(err, msg)
			op:setFailed('Upload fieled: ' .. msg)
		end}
end

-----------------------------------------------------------------------
-- Write, tar config and install
function write_config(op)
	local fname = string.format('NTC-6000_Settings_%06d.cfg', math.random(999999))
	local f = io.open('/tmp/'..fname, 'w+')
	f:write(op.record[3])
	f:close()
	removeCredentials('/tmp/'..fname)
	local openvpnkeys = file_entry_exists('/usr/local/cdcs/openvpn-keys') and 'openvpn-keys ' or ''
	local sshhostkeys = file_entry_exists('/usr/local/cdcs/ssh-hostkeys') and 'ssh-hostkeys ' or ''
	local c = string.format('cd /tmp && tar -C /usr/local/cdcs -czf - ipsec.d ' .. openvpnkeys .. sshhostkeys .. '| openssl des3 -salt -k "" | dd of=/tmp/vpn.des3 && tar czf /opt/cdcs/upload/%s.tar.gz %s vpn.des3', fname, fname)
	debug( 'conf cmd:', c )
	os.execute(c)
	local ok, output = installFile(string.format('/opt/cdcs/upload/%s.tar.gz', fname))
	os.remove('/tmp/'..fname)
	os.remove('/tmp/vpn.des3')
	if ok then
		if output:find('reboot') then
			-- reboot
			os.execute('rdb set service.system.reset 1')
		else
			op.successful = true
		end
	else
		DEBUG('installing /opt/cdcs/upload/'..fname..'.tar.gz failed: ' .. output)
		op:setFailed('install_file failed: '..output)
	end
end

-----------------------------------------------------------------------
-- Send config to c8y
function send_config()
	local f = io.popen("dbcfg_export -p '' | sed " .. removeCredentialsSedPattern, 'r')
	local c = f:read('*all')
	--DUMP( c )
	templates.update:send({c8y.device, c})
end

function read_config(op)
	send_config()
	op.successful = true
end

-----------------------------------------------------------------------
-- Plugin is starting
function start()
	-- set successful afrer reboot
	local op = c8y.plugin.startupExecutingOperations[1]
  if op then
		op.successful = true
  end
	-- check if just registred
	if c8y.device.registered then
		send_config()
	end
end

-----------------------------------------------------------------------
-- Plugin initialize
function init()
  debug("CONFIG FILE PLUGIN DEBUGGING ENABLED")
	-- c8y_Configuration
	c8y.operations:addOperation('c8y_Configuration', '$.config', write_config)
	c8y.operations:addOperation('c8y_SendConfiguration', '', read_config)
	templates.update = c8y.templates:addRequest('PUT,/inventory/managedObjects/%%,application/vnd.com.nsn.cumulocity.managedObject+json,application/vnd.com.nsn.cumulocity.managedObject+json,%%,UNSIGNED STRING,"{""c8y_Configuration"":{""config"":""%%""}}"')
	-- c8y_DownloadConfigFile / c8y_UploadConfigFile
  c8y.operations:addOperation("c8y_DownloadConfigFile","$.url,$.c8y_ConfigurationDump.id", download)
  c8y.operations:addOperation{'c8y_UploadConfigFile', '', upload}
  templates.createConfigRepoEntry = c8y.templates:addRequest('POST,/inventory/managedObjects,application/vnd.com.nsn.cumulocity.managedObject+json,application/vnd.com.nsn.cumulocity.managedObject+json,%%,STRING STRING STRING,"{""name"":""%%"",""description"":""%%"",""url"":""%%"",""type"":""c8y_ConfigurationDump""}"')
  templates.updateConfigDumpFragment = c8y.templates:addRequest('PUT,/inventory/managedObjects/%%,application/vnd.com.nsn.cumulocity.managedObject+json,,%%,UNSIGNED UNSIGNED,"{""c8y_ConfigurationDump"":{""id"":""%%""}}"')
  templates.response = c8y.templates:addResponse(',$.id,$.id')
	-- restore credentials
	if file_exists(credencialsFile) then
		DEBUG('restore credetials')
		os.execute([[grep 'service.cumulocity.connection.\(password\|username\)' ]] ..credencialsFile
			..[[ | while read k v; do rdb_set "$k" "$v"; done]])
		-- TODO:  Good place to check is server is changed (MTM-9595)
		os.remove(credencialsFile)
	end
end



-----------------------------------------------------------------------
--- XXX   Tests   XXX  ---
if c8y.config:value("xxxTests") then
exportConfig = function()
	DUMP('Make configuration dump...')
  local fileName = string.format("NTC-6200_Settings_%d_%s.cfg.tar.gz", c8y.device.id, c8y.strftime('%Y%m%d'))
	os.execute('echo -n 12345 > ' .. fileName)
  return fileName 
end
installFile = function(fileName)
	DUMP('\nINSTALLING "' .. (filename or '') .. '"...\n')
	return true
end
reboot = function()
  DUMP('\n\n        REBOOT !!!   \n\n')
end
end   ---  Tests


